package edu.berkeley.qfat;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.media.opengl.*;
import javax.media.opengl.glu.*;
import com.sun.opengl.util.*;
import java.util.*;
import edu.berkeley.qfat.geom.*;
import edu.berkeley.qfat.geom.Point;

public class MeshViewer implements GLEventListener, MouseListener, MouseMotionListener, KeyListener, MouseWheelListener  {

    public Mesh tile = new Mesh(false);
    public Mesh goal = new Mesh(false);
    public Matrix[] translations;
    public Mesh.Vertex[] points;


    public boolean tileon = true;
    public boolean tilemeshon = false;
    public boolean goalon = false;
    public boolean anneal = false;
    public boolean neighbors = false;

    public int breaks = 0;
    boolean alt = false;
    boolean shift = false;
    boolean control = false;

    public void mouseWheelMoved(MouseWheelEvent e) {
        tz -= e.getWheelRotation();
    }

    public void keyTyped(KeyEvent e)  { }
    public void keyPressed(KeyEvent e)  {
        switch(e.getKeyCode()) {
            case KeyEvent.VK_CONTROL: control = true; break;
            case KeyEvent.VK_ALT: alt = true; break;
            case KeyEvent.VK_SHIFT: shift = true; break;
            case KeyEvent.VK_SPACE: breaks++; break;
            case KeyEvent.VK_D: dump(); break;
            case KeyEvent.VK_A: anneal = !anneal; break;
            case KeyEvent.VK_N: neighbors = !neighbors; break;
            case KeyEvent.VK_T: tileon = !tileon; break;
            case KeyEvent.VK_G: goalon = !goalon; break;
            case KeyEvent.VK_M: tilemeshon = !tilemeshon; break;
        }
    }
    public synchronized void dump() {
        try {
        PrintWriter pw = new PrintWriter(new OutputStreamWriter(new FileOutputStream("dump.stl")));
        pw.println("solid dump");
        for(Mesh.T t : tile) {
            Vec normal = t.norm();
            pw.println("facet normal " + normal.x + " " + normal.y + " " + normal.z);
            pw.println("  outer loop");
            for(Mesh.Vertex v : new Mesh.Vertex[] { t.v1(), t.v2(), t.v3() }) {
                pw.println("    vertex " + v.p.x + " " + v.p.y + " " + v.p.z);
            }
            pw.println("  endloop");
            pw.println("endfacet");
        }
        pw.println("endsolid dump");
        pw.flush();
        pw.close();
        } catch (Exception e) { throw new RuntimeException(e); }
    }
    public void keyReleased(KeyEvent e) {
        switch(e.getKeyCode()) {
            case KeyEvent.VK_CONTROL: control = false; break;
            case KeyEvent.VK_ALT: alt = false; break;
            case KeyEvent.VK_SHIFT: shift = false; break;
        }
    }

    public void mouseClicked(MouseEvent e) { }
    public void mouseEntered(MouseEvent e) { }
    public void mouseExited(MouseEvent e) { }
    public void mousePressed(MouseEvent e) { }
    public void mouseReleased(MouseEvent e) { }

    int mousex;
    int mousey;
    public void mouseMoved(MouseEvent e) {
        mousex = e.getX();
        mousey = e.getY();
    }

    float tx = 0;
    float ty = 0;
    float tz = 0;
    float anglex = 0;
    float angley = 0;
    public void mouseDragged(MouseEvent e) {
        if (shift) {
            tx += (mousex - e.getX())/(float)20;
            ty += (mousey - e.getY())/(float)20;
        } else {
            anglex -= mousex - e.getX();
            angley += mousey - e.getY();
        }
        mousex = e.getX();
        mousey = e.getY();
    }

    /**
     * Take care of initialization here.
     */
    public void init(GLAutoDrawable gld) {
        GL gl = gld.getGL();
        gl.glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
        gl.glViewport(0, 0, 500, 300);
        gl.glEnable(GL.GL_DEPTH_TEST);
        gl.glClearDepth(1.0);
        gl.glDepthFunc(GL.GL_LEQUAL);
        gl.glMatrixMode(GL.GL_PROJECTION);
        gl.glLoadIdentity();
        gl.glMatrixMode(GL.GL_MODELVIEW);

        float mat_specular[] = { 0.5f, 0.5f, 0.5f, 0.5f };
        float mat_shininess[] = { 50.0f };
        gl.glShadeModel(GL.GL_SMOOTH);
        //gl.glMaterialfv(GL.GL_FRONT, GL.GL_DIFFUSE, mat_specular, 0);
        //gl.glMaterialfv(GL.GL_FRONT, GL.GL_SPECULAR, mat_specular, 0);  
        //gl.glMaterialfv(GL.GL_FRONT, GL.GL_AMBIENT, new float[] { 0.3f, 0.3f, 0.3f, 0.3f }, 0);  
        //gl.glMaterialfv(GL.GL_FRONT, GL.GL_SHININESS, mat_shininess, 0);
        gl.glLightfv(GL.GL_LIGHT0, GL.GL_POSITION, new float[] { 1.0f,    4.0f,  -10.0f, 0.0f }, 0);
        gl.glLightfv(GL.GL_LIGHT1, GL.GL_POSITION, new float[] { -10.0f, 10.0f,   10.0f, 0.0f }, 0);
        gl.glLightfv(GL.GL_LIGHT2, GL.GL_POSITION, new float[] { 10.0f, -10.0f,   10.0f, 0.0f }, 0);
        gl.glLightfv(GL.GL_LIGHT3, GL.GL_POSITION, new float[] { 10.0f,  10.0f,  -10.0f, 0.0f }, 0);
        gl.glLightfv(GL.GL_LIGHT4, GL.GL_POSITION, new float[] { -10.0f, 10.0f,  -10.0f, 0.0f }, 0);
        gl.glLightfv(GL.GL_LIGHT5, GL.GL_POSITION, new float[] { 10.0f, -10.0f,  -10.0f, 0.0f }, 0);
        gl.glEnable(GL.GL_LIGHTING);
        gl.glEnable(GL.GL_LIGHT0);
        /*
        gl.glEnable(GL.GL_LIGHT1);
        gl.glEnable(GL.GL_LIGHT2);
        gl.glEnable(GL.GL_LIGHT3);
        gl.glEnable(GL.GL_LIGHT4);
        gl.glEnable(GL.GL_LIGHT5);
        */
        gl.glColorMaterial(GL.GL_FRONT_AND_BACK, GL.GL_AMBIENT_AND_DIFFUSE);
        gl.glEnable(GL.GL_COLOR_MATERIAL);

        display(gld);
    }

    public int temps;
    public int accepts;
    public    int vertss;
    public void reshape(GLAutoDrawable drawable, int x, int y, int width, int height) { }
    public void displayChanged(GLAutoDrawable drawable, boolean modeChanged, boolean deviceChanged) { }
    public void display(GLAutoDrawable drawable) {

        if (translations==null) return;

        glcanvas.setSize(f.getWidth(), f.getHeight() - 100);
        Graphics2D g = (Graphics2D)f.getGraphics();
        g.setColor(Color.black);
        g.fillRect(0, f.getHeight()-100, f.getWidth(), f.getHeight());
        g.setColor(Color.red);
        int top = f.getHeight()-100;
        g.drawString("temperature: "+temps, 10, 30+top);
        g.drawString("acceptance: "+accepts, 10, 50+top);
        g.drawString("vertices: "+vertss, 10, 70+top);
        g.fillRect(140, 25+top, temps, 10);
        g.fillRect(140, 45+top, accepts, 10);
        g.fillRect(140, 65+top, vertss, 10);

        GL gl = drawable.getGL();
        GLU glu = new GLU();
        gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
        gl.glPointSize(5.0f);
        gl.glLoadIdentity();
        glu.gluPerspective(50, ((float)drawable.getWidth())/drawable.getHeight(), 0.5, 10);
        glu.gluLookAt(0, 0, (tz/10)-1, 0, 0, 0, 0, 1, 0);
        gl.glRotatef(anglex/3, 0, 1, 0);
        gl.glRotatef(angley/3, 1, 0, 0);

        gl.glBegin(GL.GL_TRIANGLES);
        if (tileon)
        draw(gl, true, safeTriangles);
        if (tilemeshon)
        draw(gl, false, safeTriangles);
        gl.glEnd();

        //draw(gl, false, tile);

        gl.glBegin(GL.GL_TRIANGLES);
        gl.glColor4f((float)0.5, (float)0.5, (float)0.5, (float)0.8);
        if (goalon)
        draw(gl, false, goal);
        gl.glEnd();


        int i = 0;
        //gl.glDisable(GL.GL_DEPTH_TEST);
        gl.glColor4f(1,1,1,1);
        for(Matrix m : translations) {
            //if (v1.z==0 && v1.y==0) continue;
            i++;
            //if (i>4) continue;
            /*
            Point p = new Point(0, 0, 0).times(m);
            Vec v = new Vec(p.x, p.y, p.z);
            v = v.times((float)1.04);
            gl.glTranslatef(v.x, v.y, v.z);
            */
            if (neighbors) draw(gl, false, safeTriangles, m);
            /*
            gl.glTranslatef(-v.x, -v.y, -v.z);
            */
        }
        //gl.glEnable(GL.GL_DEPTH_TEST);
        gl.glEnable (GL.GL_LIGHTING);
    }

    protected HashSet<Mesh.T> safeTriangles = new HashSet<Mesh.T>();

    private void draw(GL gl, boolean triangles, Iterable<Mesh.T> tris) { draw(gl, triangles, tris, Matrix.ONE); }
    private void draw(GL gl, boolean triangles, Iterable<Mesh.T> tris, Matrix m) {
        float red = 0.0f;
        float green = 0.0f;
        float blue = 0.0f;
        synchronized(safeTriangles) {
        for(Mesh.T t : tris) {
            if (red < 0.15) red = 1.0f;
            if (green < 0.15) green = 1.0f;
            if (blue < 0.15) blue = 1.0f;
            red -= .09f;
            green -= .12f;
            blue -= .15f;

            if (triangles) switch(t.color/*class*/) {
                case 0: gl.glColor4f((float)0.25, (float)0.25, (float)0.75, (float)0.3); break;
                case 1: gl.glColor4f((float)0.25, (float)0.75, (float)0.25, (float)0.3); break;
                case 2: gl.glColor4f((float)0.75, (float)0.25, (float)0.25, (float)0.3); break;
                case 3: gl.glColor4f((float)0.50, (float)0.50, (float)0.50, (float)0.3); break;
                case 4: gl.glColor4f((float)0.25, (float)0.75, (float)0.75, (float)0.3); break;
                case 5: gl.glColor4f((float)0.25, (float)0.75, (float)0.75, (float)0.3); break;
                case 6: gl.glColor4f((float)0.75, (float)0.25, (float)0.75, (float)0.3); break;
            }
            /*
            if (t.e1().pair.t==null) gl.glColor4f((float)0.25, (float)0.25, (float)0.75, (float)0.3);
            else if (t.e2().pair.t==null) gl.glColor4f((float)0.25, (float)0.25, (float)0.75, (float)0.3);
            else if (t.e3().pair.t==null) gl.glColor4f((float)0.25, (float)0.25, (float)0.75, (float)0.3);
            else  gl.glColor4f((float)0.75, (float)0.25, (float)0.25, (float)0.3);
            */
            //gl.glBegin(GL.GL_LINES);

            if (triangles) {
                gl.glBegin(GL.GL_TRIANGLES);
                t.glVertices(gl);
                gl.glEnd();
            } else {

                gl.glDisable(GL.GL_LIGHTING);
                gl.glBegin(GL.GL_LINES);
                gl.glColor3f(1, 1, 1);
                m.times(t.e1().p1.p).glVertex(gl);
                m.times(t.e1().p2.p).glVertex(gl);
                m.times(t.e2().p1.p).glVertex(gl);
                m.times(t.e2().p2.p).glVertex(gl);
                m.times(t.e3().p1.p).glVertex(gl);
                m.times(t.e3().p2.p).glVertex(gl);
                gl.glEnd();
                gl.glEnable(GL.GL_LIGHTING);

            }

            Point centroid = t.centroid();
            gl.glBegin(GL.GL_LINES);
            gl.glColor3f(1, 1, 1);
            /*
            centroid.glVertex(gl);
            centroid.plus(t.norm().times(t.diameter())).glVertex(gl);
            */
            /*
            if (mesh==goal)
                for(Mesh.Vertex p : new Mesh.Vertex[] { t.v1(), t.v2(), t.v3() }) {
                gl.glDisable(GL.GL_LIGHTING);
                gl.glBegin(GL.GL_LINES);
                gl.glColor3f(1, 1, 1);
                p.p.glVertex(gl);
                //p.p.plus(p.norm().times(p.score())).glVertex(gl);
                if (p.nearest_in_other_mesh != null) p.nearest_in_other_mesh.p.glVertex(gl);
                    //tile.nearest(p).centroid().glVertex(gl);
                gl.glEnd();
                gl.glEnable(GL.GL_LIGHTING);
                }
            */
            gl.glEnd();
        }
        }
    }


    //private JTextArea ocanvas = new JTextArea();
    private Frame f;
    private GLCanvas glcanvas;
    public MeshViewer(Frame f) {
        this.f = f;
       GLCapabilities glcaps = new GLCapabilities();
        glcanvas = new GLCanvas();
        glcanvas.addGLEventListener(this);
        f.add(glcanvas, BorderLayout.CENTER);
        glcanvas.addMouseListener(this);
        glcanvas.addMouseMotionListener(this);
        glcanvas.addMouseWheelListener(this);
        glcanvas.addKeyListener(this);
    }
    public void repaint() {
        glcanvas.repaint();
    }
}