package edu.berkeley.qfat.geom;
import javax.media.opengl.*;
import java.util.*;

public abstract class HasBindingGroup {

    public BindingGroup bindingGroup;

    // know:      self   = bindingGroup[self] * master
    // know:      other  = other.bindingGroup[other] * other.bindingGroup.master
    // want:      self   = bindingMatrix * other
    // therefore: master = bindingGroup[self]^-1 * bindingMatrix * other
    // therefore:        = bindingGroup[self]^-1 * bindingMatrix * other.bindingGroup[other] * other.bindingGroup.master
    public void bindTo(Matrix bindingMatrix, HasBindingGroup other) {
        if (bindingGroup == null) bindingGroup = new BindingGroup(this);
        if (other.bindingGroup == null) other.bindingGroup = new BindingGroup(other);

        if (other.bindingGroup == this.bindingGroup) {
            if (getBindingMatrix(other).equals(bindingMatrix))
                return;
            /*
            throw new Error("rebind attempt: "+this+" and "+other+" with "
                            + bindingMatrix
                            + "; expected "
                            + getBindingMatrix(other));
            */
            Matrix vom = other.getBindingMatrix(bindingGroup.getMaster());
            Matrix v2 = bindingMatrix;
            Matrix v3 = getBindingMatrix(other);
            Matrix m = vom.inverse().times(v2.inverse()).times(v3).times(vom);
            bindingGroup.krank = bindingGroup.krank.times(m);
            /*
              System.err.println("WARNING: discarding rebind attempt: "+this+" and "+other+m);
            */
            //" with "
            //+ bindingMatrix
            //+ "; expected "
            //+ getBindingMatrix(other));
            return;
        }


        bindingMatrix =
            getBindingMatrix().inverse()
            .times(bindingMatrix)
            .times(other.getBindingMatrix());
        other.bindingGroup.merge(bindingGroup, bindingMatrix);
    }

    public int bindingGroupSize() {
        if (bindingGroup == null) return 1;
        return bindingGroup.size();
    }


    public Matrix getBindingMatrix() {
        if (bindingGroup==null) return Matrix.ONE;
        return bindingGroup.getMatrix(this);
    }

    public Matrix getBindingMatrix(HasBindingGroup other) {
        if (other==this) return Matrix.ONE;
        return bindingGroup.getMatrix(this, other);
    }

    public HasBindingGroup getMaster() {
        if (bindingGroup==null) return this;
        return bindingGroup.getMaster();
    }

    public void unbind() {
        if (bindingGroup==null) return;
        bindingGroup.unbind(this);
        bindingGroup = null;
        bindingGroupChanged(bindingGroup);
    }

    public boolean isBoundTo(HasBindingGroup t) {
        return t==this || (bindingGroup!=null && bindingGroup.contains(t));
    }

    public Iterable getBoundPeers() {
        if (bindingGroup==null) bindingGroup = new BindingGroup(this);
        return bindingGroup;
    }
    public void bindingGroupChanged(BindingGroup newBindingGroup) { }
}
